#!/usr/bin/env python
# License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>

import textwrap

# from lxml.etree import Element
from calibre import force_unicode
from calibre.ebooks.oeb.base import OEB_DOCS, OEB_STYLES, SVG, XHTML, XPNSMAP, barename, serialize
from calibre.ebooks.oeb.polish.container import OPF_NAMESPACES
from calibre.ebooks.oeb.polish.utils import guess_type
from calibre.utils.icu import sort_key


def isspace(x):
    return not x.strip('\u0009\u000a\u000c\u000d\u0020')


def pretty_xml_tree(elem, level=0, indent='  '):
    """XML beautifier, assumes that elements that have children do not have
    textual content.  Also assumes that there is no text immediately after
    closing tags. These are true for opf/ncx and container.xml files. If either
    of the assumptions are violated, there should be no data loss, but pretty
    printing won't produce optimal results."""
    if (not elem.text and len(elem) > 0) or (elem.text and isspace(elem.text)):
        elem.text = '\n' + (indent * (level + 1))
    for i, child in enumerate(elem):
        pretty_xml_tree(child, level=level + 1, indent=indent)
        if not child.tail or isspace(child.tail):
            l = level + 1
            if i == len(elem) - 1:
                l -= 1
            child.tail = '\n' + (indent * l)


def pretty_opf(root):
    # Put all dc: tags first starting with title and author. Preserve order for
    # the rest.
    def dckey(x):
        return {'title': 0, 'creator': 1}.get(barename(x.tag), 2)

    for metadata in root.xpath('//opf:metadata', namespaces=OPF_NAMESPACES):
        dc_tags = metadata.xpath('./*[namespace-uri()="{}"]'.format(OPF_NAMESPACES['dc']))
        dc_tags.sort(key=dckey)
        for x in reversed(dc_tags):
            metadata.insert(0, x)

    # Group items in the manifest
    spine_ids = root.xpath('//opf:spine/opf:itemref/@idref', namespaces=OPF_NAMESPACES)
    spine_ids = {x: i for i, x in enumerate(spine_ids)}

    def manifest_key(x):
        mt = x.get('media-type', '')
        href = x.get('href', '')
        ext = href.rpartition('.')[-1].lower()
        cat = 1000
        if mt in OEB_DOCS:
            cat = 0
        elif mt == guess_type('a.ncx'):
            cat = 1
        elif mt in OEB_STYLES:
            cat = 2
        elif mt.startswith('image/'):
            cat = 3
        elif ext in {'otf', 'ttf', 'woff', 'woff2'}:
            cat = 4
        elif mt.startswith('audio/'):
            cat = 5
        elif mt.startswith('video/'):
            cat = 6

        if cat == 0:
            i = spine_ids.get(x.get('id', None), 1000000000)
        else:
            i = sort_key(href)
        return cat, i

    for manifest in root.xpath('//opf:manifest', namespaces=OPF_NAMESPACES):
        try:
            children = sorted(manifest, key=manifest_key)
        except AttributeError:
            continue  # There are comments so don't sort since that would mess up the comments
        for x in reversed(children):
            manifest.insert(0, x)


SVG_TAG = SVG('svg')
NON_NAMESPACED_BLOCK_TAGS = (
    'address',
    'article',
    'aside',
    'audio',
    'blockquote',
    'body',
    'canvas',
    'col',
    'colgroup',
    'dd',
    'div',
    'dl',
    'dt',
    'fieldset',
    'figcaption',
    'figure',
    'footer',
    'form',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'header',
    'hgroup',
    'hr',
    'li',
    'svg',
    'noscript',
    'ol',
    'output',
    'p',
    'pre',
    'script',
    'section',
    'style',
    'table',
    'tbody',
    'td',
    'tfoot',
    'th',
    'thead',
    'tr',
    'ul',
    'video',
    'img',
)
BLOCK_TAGS = frozenset(map(XHTML, NON_NAMESPACED_BLOCK_TAGS))


def isblock(x):
    if callable(x.tag) or not x.tag:
        return True
    if x.tag in BLOCK_TAGS:
        return True
    return False


def has_only_blocks(x):
    if hasattr(x.tag, 'split') and len(x) == 0:
        # Tag with no children,
        return False
    if x.text and not isspace(x.text):
        return False
    for child in x:
        if not isblock(child) or (child.tail and not isspace(child.tail)):
            return False
    return True


def indent_for_tag(x):
    prev = x.getprevious()
    x = x.getparent().text if prev is None else prev.tail
    if not x:
        return ''
    s = x.rpartition('\n')[-1]
    return s if isspace(s) else ''


def set_indent(elem, attr, indent):
    x = getattr(elem, attr)
    if not x:
        x = indent
    else:
        lines = x.splitlines()
        if isspace(lines[-1]):
            lines[-1] = indent
        else:
            lines.append(indent)
        x = '\n'.join(lines)
    setattr(elem, attr, x)


def pretty_block(parent, level=1, indent='  '):
    """Surround block tags with blank lines and recurse into child block tags
    that contain only other block tags"""
    if not parent.text or isspace(parent.text):
        parent.text = ''
    nn = '\n' if hasattr(parent.tag, 'strip') and barename(parent.tag) in {'tr', 'td', 'th'} else '\n\n'
    parent.text = parent.text + nn + (indent * level)
    for i, child in enumerate(parent):
        if isblock(child) and has_only_blocks(child):
            pretty_block(child, level=level + 1, indent=indent)
        elif child.tag == SVG_TAG:
            pretty_xml_tree(child, level=level, indent=indent)
        l = level
        if i == len(parent) - 1:
            l -= 1
        if not child.tail or isspace(child.tail):
            child.tail = ''
        child.tail = child.tail + nn + (indent * l)


def pretty_script_or_style(container, child):
    if child.text:
        indent = indent_for_tag(child)
        if child.tag.endswith('style'):
            child.text = force_unicode(pretty_css(container, '', child.text), 'utf-8')
        child.text = textwrap.dedent(child.text)
        child.text = '\n' + '\n'.join([(indent + x) if x else '' for x in child.text.splitlines()])
        set_indent(child, 'text', indent)


def pretty_html_tree(container, root):
    root.text = '\n\n'
    for child in root:
        child.tail = '\n\n'
        if hasattr(child.tag, 'endswith') and child.tag.endswith('}head'):
            pretty_xml_tree(child)
    for body in root.findall('h:body', namespaces=XPNSMAP):
        pretty_block(body)
        # Special case the handling of a body that contains a single block tag
        # with all content. In this case we prettify the containing block tag
        # even if it has non block children.
        if (
            len(body) == 1
            and not callable(body[0].tag)
            and isblock(body[0])
            and not has_only_blocks(body[0])
            and barename(body[0].tag) not in ('pre', 'p', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6')
            and len(body[0]) > 0
        ):
            pretty_block(body[0], level=2)

    if container is not None:
        # Handle <script> and <style> tags
        for child in root.xpath('//*[local-name()="script" or local-name()="style"]'):
            pretty_script_or_style(container, child)


def fix_html(container, raw):
    "Fix any parsing errors in the HTML represented as a string in raw. Fixing is done using the HTML5 parsing algorithm."
    root = container.parse_xhtml(raw)
    return serialize(root, 'text/html')


def pretty_html(container, name, raw):
    "Pretty print the HTML represented as a string in raw"
    root = container.parse_xhtml(raw)
    pretty_html_tree(container, root)
    return serialize(root, 'text/html')


def pretty_css(container, name, raw):
    "Pretty print the CSS represented as a string in raw"
    sheet = container.parse_css(raw)
    return serialize(sheet, 'text/css')


def pretty_xml(container, name, raw):
    "Pretty print the XML represented as a string in raw. If ``name`` is the name of the OPF, extra OPF-specific prettying is performed."
    root = container.parse_xml(raw)
    if name == container.opf_name:
        pretty_opf(root)
    pretty_xml_tree(root)
    return serialize(root, 'text/xml')


def fix_all_html(container):
    "Fix any parsing errors in all HTML files in the container. Fixing is done using the HTML5 parsing algorithm."
    for name, mt in container.mime_map.items():
        if mt in OEB_DOCS:
            container.parsed(name)
            container.dirty(name)


def pretty_all(container):
    "Pretty print all HTML/CSS/XML files in the container"
    xml_types = {guess_type('a.ncx'), guess_type('a.xml'), guess_type('a.svg')}
    for name, mt in container.mime_map.items():
        prettied = False
        if mt in OEB_DOCS:
            pretty_html_tree(container, container.parsed(name))
            prettied = True
        elif mt in OEB_STYLES:
            container.parsed(name)
            prettied = True
        elif name == container.opf_name:
            root = container.parsed(name)
            pretty_opf(root)
            pretty_xml_tree(root)
            prettied = True
        elif mt in xml_types:
            pretty_xml_tree(container.parsed(name))
            prettied = True
        if prettied:
            container.dirty(name)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2014, Kovid Goyal <kovid at kovidgoyal.net>


def load_patience_module():
    from calibre_extensions import _patiencediff_c

    return _patiencediff_c


def get_sequence_matcher():
    return load_patience_module().PatienceSequenceMatcher_c



#!/usr/bin/env python
# License: GPLv3 Copyright: 2008, Kovid Goyal kovid@kovidgoyal.net

"""
Convert CBR/CBZ files to LRF.
"""

import sys


def main(args=sys.argv):
    return 0


if __name__ == '__main__':
    sys.exit(main())



# License: GPLv3 Copyright: 2009, John Schember <john@nachtimwald.com>

import os


class EreaderError(Exception):
    pass


def image_name(name, taken_names=()):
    name = os.path.basename(name)

    if len(name) > 32:
        cut = len(name) - 32
        names = name[:10]
        namee = name[10 + cut :]
        name = f'{names}{namee}.png'

    i = 0
    base_name, ext = os.path.splitext(name)
    while name in taken_names:
        i += 1
        name = f'{base_name}{i}{ext}'

    return name.ljust(32, '\x00')[:32]



#!/usr/bin/env python
# License: GPLv3 Copyright: 2015, Kovid Goyal <kovid at kovidgoyal.net>

import errno
import numbers
import os
from collections import OrderedDict, namedtuple
from functools import partial
from itertools import zip_longest
from operator import attrgetter

from calibre.constants import config_dir
from calibre.utils.localization import _
from calibre.utils.lock import ExclusiveFile

Option = namedtuple('Option', 'name default longdoc shortdoc choices')


class Choices(frozenset):
    default: str

    def __new__(cls, *args):
        self = super().__new__(cls, args)
        self.default = args[0]
        return self


raw_options = (
    _('Path to the SSL certificate file'),
    'ssl_certfile',
    None,
    None,
    _('Path to the SSL private key file'),
    'ssl_keyfile',
    None,
    None,
    _('Time (in seconds) after which an idle connection is closed'),
    'timeout',
    120.0,
    None,
    _('Time (in seconds) to wait for a response from the server when making queries'),
    'ajax_timeout',
    300.0,
    None,
    _('Total time in seconds to wait for clean shutdown'),
    'shutdown_timeout',
    5.0,
    None,
    _('Socket pre-allocation, for example, with systemd socket activation'),
    'allow_socket_preallocation',
    True,
    None,
    _('Max. size of single HTTP header (in KB)'),
    'max_header_line_size',
    8.0,
    None,
    _('Max. allowed size for files uploaded to the server (in MB)'),
    'max_request_body_size',
    500.0,
    None,
    _('Minimum size for which responses use data compression (in bytes)'),
    'compress_min_size',
    1024,
    None,
    _('Number of worker threads used to process requests'),
    'worker_count',
    10,
    None,
    _('Maximum number of worker processes'),
    'max_jobs',
    0,
    _(
        'Worker processes are launched as needed and used for large jobs such as preparing'
        ' a book for viewing, adding books, converting, etc. Normally, the max.'
        ' number of such processes is based on the number of CPU cores. You can'
        ' control it by this setting.'
    ),
    _('Maximum time for worker processes'),
    'max_job_time',
    60,
    _('Maximum amount of time worker processes are allowed to run (in minutes). Set to zero for no limit.'),
    _('The port on which to listen for connections'),
    'port',
    8080,
    None,
    _('A prefix to prepend to all URLs'),
    'url_prefix',
    None,
    _('Useful if you wish to run this server behind a reverse proxy. For example use, /calibre as the URL prefix.'),
    _('Number of books to show in a single page'),
    'num_per_page',
    50,
    _('The number of books to show in a single page in the browser.'),
    _('Advertise OPDS feeds via BonJour'),
    'use_bonjour',
    True,
    _('Advertise the OPDS feeds via the BonJour service, so that OPDS based reading apps can detect and connect to the server automatically.'),
    _('Maximum number of books in OPDS feeds'),
    'max_opds_items',
    30,
    _('The maximum number of books that the server will return in a single OPDS acquisition feed.'),
    _('Maximum number of ungrouped items in OPDS feeds'),
    'max_opds_ungrouped_items',
    100,
    _('Group items in categories such as author/tags by first letter when there are more than this number of items. Set to zero to disable.'),
    _('The interface on which to listen for connections'),
    'listen_on',
    None,
    _(
        'The default is to listen on all available IPv6 and IPv4 interfaces. You can change this to, for'
        ' example, "127.0.0.1" to only listen for IPv4 connections from the local machine, or'
        ' to "0.0.0.0" to listen to all incoming IPv4 connections.'
    ),
    _('Fallback to auto-detected interface'),
    'fallback_to_detected_interface',
    True,
    _(
        'If for some reason the server is unable to bind to the interface specified in'
        ' the listen_on option, then it will try to detect an interface that connects'
        ' to the outside world and bind to that.'
    ),
    _('Zero copy file transfers for increased performance'),
    'use_sendfile',
    True,
    _(
        'This will use zero-copy in-kernel transfers when sending files over the network,'
        ' increasing performance. However, it can cause corrupted file transfers on some'
        ' broken filesystems. If you experience corrupted file transfers, turn it off.'
    ),
    _('Max. log file size (in MB)'),
    'max_log_size',
    20,
    _(
        'The maximum size of log files, generated by the server. When the log becomes larger'
        ' than this size, it is automatically rotated. Set to zero to disable log rotation.'
    ),
    _('Log HTTP 404 (Not Found) requests'),
    'log_not_found',
    True,
    _(
        'Normally, the server logs all HTTP requests for resources that are not found.'
        ' This can generate a lot of log spam, if your server is targeted by bots.'
        ' Use this option to turn it off.'
    ),
    _('Password based authentication to access the server'),
    'auth',
    False,
    _('Normally, the server is unrestricted, allowing anyone to access it. You can restrict access to predefined users with this option.'),
    _('Allow un-authenticated local connections to make changes'),
    'local_write',
    False,
    _(
        'Normally, if you do not turn on authentication, the server operates in'
        ' read-only mode, so as to not allow anonymous users to make changes to your'
        ' calibre libraries. This option allows anybody connecting from the same'
        ' computer as the server is running on to make changes. This is useful'
        ' if you want to run the server without authentication but still'
        ' use calibredb to make changes to your calibre libraries. Note that'
        ' turning on this option means any program running on the computer'
        ' can make changes to your calibre libraries. In particular, the Firefox web browser'
        ' allows arbitrary websites to do so as well.'
    ),
    _('Allow un-authenticated connections from specific IP addresses to make changes'),
    'trusted_ips',
    None,
    _(
        'Normally, if you do not turn on authentication, the server operates in'
        ' read-only mode, so as to not allow anonymous users to make changes to your'
        ' calibre libraries. This option allows anybody connecting from the specified'
        ' IP addresses to make changes. Must be a comma separated list of address or network specifications.'
        ' This is useful if you want to run the server without authentication but still'
        ' use calibredb to make changes to your calibre libraries. Note that'
        ' turning on this option means anyone connecting from the specified IP addresses'
        ' can make changes to your calibre libraries.'
    ),
    _('Path to user database'),
    'userdb',
    None,
    _('Path to a file in which to store the user and password information. Normally a file in the calibre configuration folder is used.'),
    _('Choose the type of authentication used'),
    'auth_mode',
    Choices('auto', 'basic', 'digest'),
    _(
        'Set the HTTP authentication mode used by the server. Set to "basic" if you are'
        ' putting this server behind an SSL proxy. Otherwise, leave it as "auto", which'
        ' will use "basic" if SSL is configured otherwise it will use "digest".'
    ),
    _('Ban IP addresses that have repeated login failures'),
    'ban_for',
    0,
    _(
        'Temporarily bans access for IP addresses that have repeated login failures for the'
        ' specified number of minutes. Useful to prevent attempts at guessing passwords. If'
        ' set to zero, no banning is done.'
    ),
    _('Number of login failures for ban'),
    'ban_after',
    5,
    _('The number of login failures after which an IP address is banned'),
    _('Ignored user-defined metadata fields'),
    'ignored_fields',
    None,
    _(
        'Comma separated list of user-defined metadata fields that will not be displayed by the Content server in the /opds and /mobile views. For example: {}'
    ).format('my_rating,my_tags'),
    _('Restrict displayed user-defined fields'),
    'displayed_fields',
    None,
    _(
        'Comma separated list of user-defined metadata fields that will be displayed'
        ' by the Content server in the /opds and /mobile views. If you specify this'
        ' option, any fields not in this list will not be displayed. For example: {}'
    ).format('my_rating,my_tags'),
    _('Choose the default book list mode'),
    'book_list_mode',
    Choices('cover_grid', 'details_list', 'custom_list'),
    _(
        'Set the default book list mode that will be used for new users. Individual users'
        ' can override the default in their own settings. The default is to use a cover grid.'
    ),
)
assert len(raw_options) % 4 == 0

_options_list: list[Option] = []


def grouper(n, iterable, fillvalue=None):
    "grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


for shortdoc, name, default, doc in grouper(4, raw_options):
    choices = None
    if isinstance(default, Choices):
        choices = sorted(default)
        default = default.default
    _options_list.append(Option(name, default, doc, shortdoc, choices))
options: OrderedDict[str, Option] = OrderedDict([(o.name, o) for o in sorted(_options_list, key=attrgetter('name'))])
del raw_options, _options_list


class Options:
    __slots__ = tuple(name for name in options)

    # Type annotations for all options (for static analysis)
    ssl_certfile: str | None
    ssl_keyfile: str | None
    timeout: float
    ajax_timeout: float
    shutdown_timeout: float
    allow_socket_preallocation: bool
    max_header_line_size: float
    max_request_body_size: float
    compress_min_size: int
    worker_count: int
    max_jobs: int
    max_job_time: int
    port: int
    url_prefix: str | None
    num_per_page: int
    use_bonjour: bool
    max_opds_items: int
    max_opds_ungrouped_items: int
    listen_on: str | None
    fallback_to_detected_interface: bool
    use_sendfile: bool
    max_log_size: int
    log_not_found: bool
    auth: bool
    local_write: bool
    trusted_ips: str | None
    userdb: str | None
    auth_mode: str
    ban_for: int
    ban_after: int
    ignored_fields: str | None
    displayed_fields: str | None
    book_list_mode: str

    def __init__(self, **kwargs):
        for opt in options.values():
            setattr(self, opt.name, kwargs.get(opt.name, opt.default))


def opt_to_cli_help(opt):
    ans = opt.shortdoc
    if not ans.endswith('.'):
        ans += '.'
    if opt.longdoc:
        ans += '\n\t' + opt.longdoc
    return ans


def bool_callback(option, opt_str, value, parser, *args, **kwargs):
    setattr(parser.values, option.dest, opt_str.startswith('--enable-'))


def boolean_option(add_option, opt):
    name = opt.name.replace('_', '-')
    help = opt_to_cli_help(opt)
    help += '\n' + (_('By default, this option is enabled.') if opt.default else _('By default, this option is disabled.'))
    add_option('--enable-' + name, '--disable-' + name, action='callback', callback=bool_callback, help=help)


def opts_to_parser(usage):
    from calibre.utils.config import OptionParser

    parser = OptionParser(usage)
    for opt in options.values():
        add_option = partial(parser.add_option, dest=opt.name, help=opt_to_cli_help(opt), default=opt.default)
        if opt.default is True or opt.default is False:
            boolean_option(add_option, opt)
        elif opt.choices:
            name = '--' + opt.name.replace('_', '-')
            add_option(name, choices=opt.choices)
        else:
            name = '--' + opt.name.replace('_', '-')
            otype = 'string'
            if isinstance(opt.default, numbers.Number):
                otype = type(opt.default).__name__
            add_option(name, type=otype)

    return parser


DEFAULT_CONFIG = os.path.join(config_dir, 'server-config.txt')


def parse_config_file(path=DEFAULT_CONFIG):
    try:
        with ExclusiveFile(path) as f:
            raw = f.read().decode('utf-8')
    except OSError as err:
        if err.errno != errno.ENOENT:
            raise
        raw = ''
    ans = {}
    for line in raw.splitlines():
        line = line.strip()
        if line.startswith('#'):
            continue
        key, rest = line.partition(' ')[::2]
        opt = options.get(key)
        if opt is None:
            continue
        val = rest
        if isinstance(opt.default, bool):
            val = val.lower() in ('true', 'yes', 'y')
        elif isinstance(opt.default, numbers.Number):
            try:
                val = type(opt.default)(rest)
            except Exception:
                raise ValueError(f'The value for {key}: {rest} is not a valid number')
        elif opt.choices:
            if rest not in opt.choices:
                raise ValueError(f'The value for {key}: {rest} is not valid')
        ans[key] = val
    return Options(**ans)


def write_config_file(opts, path=DEFAULT_CONFIG):
    changed = {name: getattr(opts, name) for name in options if getattr(opts, name) != options[name].default}
    lines = []
    for name in sorted(changed):
        o = options[name]
        lines.append('# ' + o.shortdoc)
        if o.longdoc:
            lines.append('# ' + o.longdoc)
        lines.append(f'{name} {changed[name]}')
    raw = '\n'.join(lines).encode('utf-8')
    with ExclusiveFile(path) as f:
        f.truncate()
        f.write(raw)


_server_config_cache: Options | None = None


def server_config(refresh=False):
    global _server_config_cache
    if refresh or _server_config_cache is None:
        _server_config_cache = parse_config_file()
    return _server_config_cache


def change_settings(**kwds):
    global _server_config_cache
    new_opts: dict[str, object] = {}
    opts = server_config()
    for name in options:
        if name in kwds:
            new_opts[name] = kwds[name]
        else:
            new_opts[name] = getattr(opts, name)
    new_opts_obj = Options(**new_opts)
    _server_config_cache = new_opts_obj
    write_config_file(new_opts_obj)



# License: GPLv3 Copyright: 2011, John Schember <john@nachtimwald.com>

from functools import partial

from qt.core import QIcon, QMenu, QStyledItemDelegate, Qt, QTreeView, pyqtSignal

from calibre import fit_image
from calibre.gui2 import empty_index
from calibre.gui2.metadata.single_download import RichTextDelegate
from calibre.gui2.store.search.models import Matches
from calibre.utils.localization import _


class ImageDelegate(QStyledItemDelegate):
    def paint(self, painter, option, index):
        QStyledItemDelegate.paint(self, painter, option, empty_index)
        img = index.data(Qt.ItemDataRole.DecorationRole)
        if img:
            h = option.rect.height() - 4
            w = option.rect.width()
            if isinstance(img, QIcon):
                img = img.pixmap(h - 4, h - 4)
                dpr = img.devicePixelRatio()
            else:
                dpr = img.devicePixelRatio()
                scaled, nw, nh = fit_image(img.width(), img.height(), w, h)
                if scaled:
                    img = img.scaled(
                        int(nw * dpr),
                        int(nh * dpr),
                        Qt.AspectRatioMode.IgnoreAspectRatio,
                        Qt.TransformationMode.SmoothTransformation,
                    )
            iw, ih = int(img.width() / dpr), int(img.height() / dpr)
            dx, dy = (option.rect.width() - iw) // 2, (option.rect.height() - ih) // 2
            painter.drawPixmap(option.rect.adjusted(dx, dy, -dx, -dy), img)


class ResultsView(QTreeView):
    download_requested = pyqtSignal(object)
    open_requested = pyqtSignal(object)

    def __init__(self, *args, **kwargs):
        QTreeView.__init__(self, *args, **kwargs)

        self._model = Matches()
        self.setModel(self._model)

        self.rt_delegate = RichTextDelegate(self)
        self.img_delegate = ImageDelegate(self)

        for i in self._model.HTML_COLS:
            self.setItemDelegateForColumn(i, self.rt_delegate)
        for i in self._model.IMG_COLS:
            self.setItemDelegateForColumn(i, self.img_delegate)

    def contextMenuEvent(self, a0):
        index = self.indexAt(a0.pos())

        if not index.isValid():
            return

        _m = self.model()
        assert isinstance(_m, Matches)
        result = _m.get_result(index)

        menu = QMenu(self)
        da = menu.addAction(_('Download...'), partial(self.download_requested.emit, result))
        assert da is not None
        if not result.downloads:
            da.setEnabled(False)
        menu.addSeparator()
        menu.addAction(_('Show in store'), partial(self.open_requested.emit, result))
        menu.exec(a0.globalPos())



#!/usr/bin/env python
# License: GPLv3 Copyright: 2014, Kovid Goyal <kovid at kovidgoyal.net>

import re

import pygments.unistring as uni
from pygments.lexer import RegexLexer, default, include
from pygments.token import Comment, Keyword, Name, Number, Operator, Punctuation, String, Text

from calibre.gui2.tweak_book.editor.syntax.pygments_highlighter import create_highlighter
from calibre.utils.resources import get_path as P

JS_IDENT_START = '(?:[$_' + uni.combine('Lu', 'Ll', 'Lt', 'Lm', 'Lo', 'Nl') + ']|\\\\u[a-fA-F0-9]{4})'
JS_IDENT_PART = '(?:[$' + uni.combine('Lu', 'Ll', 'Lt', 'Lm', 'Lo', 'Nl', 'Mn', 'Mc', 'Nd', 'Pc') + '\u200c\u200d]|\\\\u[a-fA-F0-9]{4})'
JS_IDENT = JS_IDENT_START + '(?:' + JS_IDENT_PART + ')*'


class JavascriptLexer(RegexLexer):
    """
    For JavaScript source code. This is based on the pygments JS highlighter,
    but that does not handle multi-line comments in streaming mode, so we had to
    modify it.
    """

    flags = re.UNICODE | re.MULTILINE

    tokens = {
        'commentsandwhitespace': [
            (r'\s+', Text),
            (r'<!--', Comment),
            (r'//.*?$', Comment.Single),
            (r'/\*', Comment.Multiline, 'comment'),
        ],
        'comment': [
            (r'[^*/]+', Comment.Multiline),
            (r'\*/', Comment.Multiline, '#pop'),
            (r'[*/]', Comment.Multiline),
        ],
        'slashstartsregex': [
            include('commentsandwhitespace'),
            (
                (
                    r'/(\\.|[^[/\\\n]|\[(\\.|[^\]\\\n])*])+/'
                    r'([gim]+\b|\B)'
                ),
                String.Regex,
                '#pop',
            ),
            (r'(?=/)', Text, ('#pop', 'badregex')),
            default('#pop'),
        ],
        'badregex': [(r'\n', Text, '#pop')],
        'root': [
            (r'\A#! ?/.*?\n', Comment),  # shebang lines are recognized by node.js
            (r'^(?=\s|/|<!--)', Text, 'slashstartsregex'),
            include('commentsandwhitespace'),
            (
                (
                    r'\+\+|--|~|&&|\?|:|\|\||\\(?=\n)|'
                    r'(<<|>>>?|==?|!=?|[-<>+*%&|^/])=?'
                ),
                Operator,
                'slashstartsregex',
            ),
            (r'[{(\[;,]', Punctuation, 'slashstartsregex'),
            (r'[})\].]', Punctuation),
            (
                (
                    r'(for|in|while|do|break|return|continue|switch|case|default|if|else|'
                    r'throw|try|catch|finally|new|delete|typeof|instanceof|void|yield|'
                    r'this)\b'
                ),
                Keyword,
                'slashstartsregex',
            ),
            (r'(var|let|with|function)\b', Keyword.Declaration, 'slashstartsregex'),
            (
                (
                    r'(abstract|boolean|byte|char|class|const|debugger|double|enum|export|'
                    r'extends|final|float|goto|implements|import|int|interface|long|native|'
                    r'package|private|protected|public|short|static|super|synchronized|throws|'
                    r'transient|volatile)\b'
                ),
                Keyword.Reserved,
            ),
            (r'(true|false|null|NaN|Infinity|undefined)\b', Keyword.Constant),
            (
                (
                    r'(Array|Boolean|Date|Error|Function|Math|netscape|'
                    r'Number|Object|Packages|RegExp|String|sun|decodeURI|'
                    r'decodeURIComponent|encodeURI|encodeURIComponent|'
                    r'Error|eval|isFinite|isNaN|parseFloat|parseInt|document|this|'
                    r'window)\b'
                ),
                Name.Builtin,
            ),
            (JS_IDENT, Name.Other),
            (r'[0-9][0-9]*\.[0-9]+([eE][0-9]+)?[fd]?', Number.Float),
            (r'0x[0-9a-fA-F]+', Number.Hex),
            (r'[0-9]+', Number.Integer),
            (r'"(\\\\|\\"|[^"])*"', String.Double),
            (r"'(\\\\|\\'|[^'])*'", String.Single),
        ],
    }


Highlighter = create_highlighter('JavascriptHighlighter', JavascriptLexer)

if __name__ == '__main__':
    from calibre.gui2.tweak_book.editor.widget import launch_editor

    launch_editor(P('viewer.js'), syntax='javascript')



#!/usr/bin/env python
# License: BSD Copyright: 2011, Leigh Parry <leighparry@blueyonder.co.uk>

"""
PyTextile

A Humane Web Text Generator
"""

# Last upstream version basis
# __version__ = '2.1.4'
# __date__ = '2009/12/04'

import re
import uuid
from urllib.parse import urlparse
from urllib.request import urlopen

from calibre.utils.smartypants import smartyPants


def _normalize_newlines(string):
    out = re.sub(r'\r\n', '\n', string)
    out = re.sub(r'\n{3,}', '\n\n', out)
    out = re.sub(r'\n\s*\n', '\n\n', out)
    out = re.sub(r'"$', '" ', out)
    return out


def getimagesize(url):
    '''
    Attempts to determine an image's width and height, and returns a string
    suitable for use in an <img> tag, or None in case of failure.
    Requires that PIL is installed.

    >>> getimagesize("http://www.google.com/intl/en_ALL/images/logo.gif")
    ... #doctest: +ELLIPSIS, +SKIP
    'width="..." height="..."'

    '''

    from PIL import ImageFile

    try:
        p = ImageFile.Parser()
        f = urlopen(url)
        while True:
            s = f.read(1024)
            if not s:
                break
            p.feed(s)
            if p.image:
                return f'width="{p.image.size[0]}" height="{p.image.size[1]}"'
    except OSError, ValueError:
        return None


class Textile:
    hlgn = r'(?:\<(?!>)|(?<!<)\>|\<\>|\=|[()]+(?! ))'
    vlgn = r'[\-^~]'
    clas = r'(?:\([^)]+\))'
    lnge = r'(?:\[[^\]]+\])'
    styl = r'(?:\{[^}]+\})'
    cspn = r'(?:\\\d+)'
    rspn = r'(?:\/\d+)'
    a = rf'(?:{hlgn}|{vlgn})*'
    s = rf'(?:{cspn}|{rspn})*'
    c = r'(?:{})*'.format('|'.join([clas, styl, lnge, hlgn]))

    pnct = r'[-!"#$%&()*+,/:;<=>?@\'\[\\\]\.^_`{|}~]'
    # urlch = r'[\w"$\-_.+!*\'(),";/?:@=&%#{}|\\^~\[\]`]'
    urlch = r'[\w"$\-_.+*\'(),";\/?:@=&%#{}|\\^~\[\]`]'

    url_schemes = ('http', 'https', 'ftp', 'mailto')

    btag = ('bq', 'bc', 'notextile', 'pre', 'h[1-6]', r'fn\d+', 'p')
    btag_lite = ('bq', 'bc', 'p')

    macro_defaults = [
        (re.compile(r'{(c\||\|c)}'), r'&#162;'),  # cent
        (re.compile(r'{(L-|-L)}'), r'&#163;'),  # pound
        (re.compile(r'{(Y=|=Y)}'), r'&#165;'),  # yen
        (re.compile(r'{\(c\)}'), r'&#169;'),  # copyright
        (re.compile(r'{\(r\)}'), r'&#174;'),  # registered
        (re.compile(r'{(\+_|_\+)}'), r'&#177;'),  # plus-minus
        (re.compile(r'{1/4}'), r'&#188;'),  # quarter
        (re.compile(r'{1/2}'), r'&#189;'),  # half
        (re.compile(r'{3/4}'), r'&#190;'),  # three-quarter
        (re.compile(r'{(A`|`A)}'), r'&#192;'),  # A-acute
        (re.compile(r'{(A\'|\'A)}'), r'&#193;'),  # A-grave
        (re.compile(r'{(A\^|\^A)}'), r'&#194;'),  # A-circumflex
        (re.compile(r'{(A~|~A)}'), r'&#195;'),  # A-tilde
        (re.compile(r'{(A\"|\"A)}'), r'&#196;'),  # A-diaeresis
        (re.compile(r'{(Ao|oA)}'), r'&#197;'),  # A-ring
        (re.compile(r'{(AE)}'), r'&#198;'),  # AE
        (re.compile(r'{(C,|,C)}'), r'&#199;'),  # C-cedilla
        (re.compile(r'{(E`|`E)}'), r'&#200;'),  # E-acute
        (re.compile(r'{(E\'|\'E)}'), r'&#201;'),  # E-grave
        (re.compile(r'{(E\^|\^E)}'), r'&#202;'),  # E-circumflex
        (re.compile(r'{(E\"|\"E)}'), r'&#203;'),  # E-diaeresis
        (re.compile(r'{(I`|`I)}'), r'&#204;'),  # I-acute
        (re.compile(r'{(I\'|\'I)}'), r'&#205;'),  # I-grave
        (re.compile(r'{(I\^|\^I)}'), r'&#206;'),  # I-circumflex
        (re.compile(r'{(I\"|\"I)}'), r'&#207;'),  # I-diaeresis
        (re.compile(r'{(D-|-D)}'), r'&#208;'),  # ETH
        (re.compile(r'{(N~|~N)}'), r'&#209;'),  # N-tilde
        (re.compile(r'{(O`|`O)}'), r'&#210;'),  # O-acute
        (re.compile(r'{(O\'|\'O)}'), r'&#211;'),  # O-grave
        (re.compile(r'{(O\^|\^O)}'), r'&#212;'),  # O-circumflex
        (re.compile(r'{(O~|~O)}'), r'&#213;'),  # O-tilde
        (re.compile(r'{(O\"|\"O)}'), r'&#214;'),  # O-diaeresis
        (re.compile(r'{x}'), r'&#215;'),  # dimension
        (re.compile(r'{(O\/|\/O)}'), r'&#216;'),  # O-slash
        (re.compile(r'{(U`|`U)}'), r'&#217;'),  # U-acute
        (re.compile(r'{(U\'|\'U)}'), r'&#218;'),  # U-grave
        (re.compile(r'{(U\^|\^U)}'), r'&#219;'),  # U-circumflex
        (re.compile(r'{(U\"|\"U)}'), r'&#220;'),  # U-diaeresis
        (re.compile(r'{(Y\'|\'Y)}'), r'&#221;'),  # Y-grave
        (re.compile(r'{sz}'), r'&szlig;'),  # sharp-s
        (re.compile(r'{(a`|`a)}'), r'&#224;'),  # a-grave
        (re.compile(r'{(a\'|\'a)}'), r'&#225;'),  # a-acute
        (re.compile(r'{(a\^|\^a)}'), r'&#226;'),  # a-circumflex
        (re.compile(r'{(a~|~a)}'), r'&#227;'),  # a-tilde
        (re.compile(r'{(a\"|\"a)}'), r'&#228;'),  # a-diaeresis
        (re.compile(r'{(ao|oa)}'), r'&#229;'),  # a-ring
        (re.compile(r'{ae}'), r'&#230;'),  # ae
        (re.compile(r'{(c,|,c)}'), r'&#231;'),  # c-cedilla
        (re.compile(r'{(e`|`e)}'), r'&#232;'),  # e-grave
        (re.compile(r'{(e\'|\'e)}'), r'&#233;'),  # e-acute
        (re.compile(r'{(e\^|\^e)}'), r'&#234;'),  # e-circumflex
        (re.compile(r'{(e\"|\"e)}'), r'&#235;'),  # e-diaeresis
        (re.compile(r'{(i`|`i)}'), r'&#236;'),  # i-grave
        (re.compile(r'{(i\'|\'i)}'), r'&#237;'),  # i-acute
        (re.compile(r'{(i\^|\^i)}'), r'&#238;'),  # i-circumflex
        (re.compile(r'{(i\"|\"i)}'), r'&#239;'),  # i-diaeresis
        (re.compile(r'{(d-|-d)}'), r'&#240;'),  # eth
        (re.compile(r'{(n~|~n)}'), r'&#241;'),  # n-tilde
        (re.compile(r'{(o`|`o)}'), r'&#242;'),  # o-grave
        (re.compile(r'{(o\'|\'o)}'), r'&#243;'),  # o-acute
        (re.compile(r'{(o\^|\^o)}'), r'&#244;'),  # o-circumflex
        (re.compile(r'{(o~|~o)}'), r'&#245;'),  # o-tilde
        (re.compile(r'{(o\"|\"o)}'), r'&#246;'),  # o-diaeresis
        (re.compile(r'{(o\/|\/o)}'), r'&#248;'),  # o-stroke
        (re.compile(r'{(u`|`u)}'), r'&#249;'),  # u-grave
        (re.compile(r'{(u\'|\'u)}'), r'&#250;'),  # u-acute
        (re.compile(r'{(u\^|\^u)}'), r'&#251;'),  # u-circumflex
        (re.compile(r'{(u\"|\"u)}'), r'&#252;'),  # u-diaeresis
        (re.compile(r'{(y\'|\'y)}'), r'&#253;'),  # y-acute
        (re.compile(r'{(y\"|\"y)}'), r'&#255;'),  # y-diaeresis
        (re.compile(r'{(C\ˇ|\ˇC)}'), r'&#268;'),  # C-caron
        (re.compile(r'{(c\ˇ|\ˇc)}'), r'&#269;'),  # c-caron
        (re.compile(r'{(D\ˇ|\ˇD)}'), r'&#270;'),  # D-caron
        (re.compile(r'{(d\ˇ|\ˇd)}'), r'&#271;'),  # d-caron
        (re.compile(r'{(E\ˇ|\ˇE)}'), r'&#282;'),  # E-caron
        (re.compile(r'{(e\ˇ|\ˇe)}'), r'&#283;'),  # e-caron
        (re.compile(r'{(L\'|\'L)}'), r'&#313;'),  # L-acute
        (re.compile(r'{(l\'|\'l)}'), r'&#314;'),  # l-acute
        (re.compile(r'{(L\ˇ|\ˇL)}'), r'&#317;'),  # L-caron
        (re.compile(r'{(l\ˇ|\ˇl)}'), r'&#318;'),  # l-caron
        (re.compile(r'{(N\ˇ|\ˇN)}'), r'&#327;'),  # N-caron
        (re.compile(r'{(n\ˇ|\ˇn)}'), r'&#328;'),  # n-caron
        (re.compile(r'{OE}'), r'&#338;'),  # OE
        (re.compile(r'{oe}'), r'&#339;'),  # oe
        (re.compile(r'{(R\'|\'R)}'), r'&#340;'),  # R-acute
        (re.compile(r'{(r\'|\'r)}'), r'&#341;'),  # r-acute
        (re.compile(r'{(R\ˇ|\ˇR)}'), r'&#344;'),  # R-caron
        (re.compile(r'{(r\ˇ|\ˇr)}'), r'&#345;'),  # r-caron
        (re.compile(r'{(S\^|\^S)}'), r'&#348;'),  # S-circumflex
        (re.compile(r'{(s\^|\^s)}'), r'&#349;'),  # s-circumflex
        (re.compile(r'{(S\ˇ|\ˇS)}'), r'&#352;'),  # S-caron
        (re.compile(r'{(s\ˇ|\ˇs)}'), r'&#353;'),  # s-caron
        (re.compile(r'{(T\ˇ|\ˇT)}'), r'&#356;'),  # T-caron
        (re.compile(r'{(t\ˇ|\ˇt)}'), r'&#357;'),  # t-caron
        (re.compile(r'{(U\°|\°U)}'), r'&#366;'),  # U-ring
        (re.compile(r'{(u\°|\°u)}'), r'&#367;'),  # u-ring
        (re.compile(r'{(Z\ˇ|\ˇZ)}'), r'&#381;'),  # Z-caron
        (re.compile(r'{(z\ˇ|\ˇz)}'), r'&#382;'),  # z-caron
        (re.compile(r'{\*}'), r'&#8226;'),  # bullet
        (re.compile(r'{Fr}'), r'&#8355;'),  # Franc
        (re.compile(r'{(L=|=L)}'), r'&#8356;'),  # Lira
        (re.compile(r'{Rs}'), r'&#8360;'),  # Rupee
        (re.compile(r'{(C=|=C)}'), r'&#8364;'),  # euro
        (re.compile(r'{tm}'), r'&#8482;'),  # trademark
        (re.compile(r'{spades?}'), r'&#9824;'),  # spade
        (re.compile(r'{clubs?}'), r'&#9827;'),  # club
        (re.compile(r'{hearts?}'), r'&#9829;'),  # heart
        (re.compile(r'{diam(onds?|s)}'), r'&#9830;'),  # diamond
        (re.compile(r'{"}'), r'&#34;'),  # double-quote
        (re.compile(r"{'}"), r'&#39;'),  # single-quote
        (re.compile(r"{(’|'/|/')}"), r'&#8217;'),  # closing-single-quote - apostrophe
        (re.compile(r"{(‘|\\'|'\\)}"), r'&#8216;'),  # opening-single-quote
        (re.compile(r'{(”|"/|/")}'), r'&#8221;'),  # closing-double-quote
        (re.compile(r'{(“|\\"|"\\)}'), r'&#8220;'),  # opening-double-quote
    ]
    glyph_defaults = [
        (re.compile(r'(\d+\'?\"?)( ?)x( ?)(?=\d+)'), r'\1\2&#215;\3'),  # dimension sign
        (re.compile(r'(\d+)\'(\s)', re.I), r'\1&#8242;\2'),  # prime
        (re.compile(r'(\d+)\"(\s)', re.I), r'\1&#8243;\2'),  # prime-double
        (
            re.compile(r'\b([A-Z][A-Z0-9]{2,})\b(?:[(]([^)]*)[)])'),
            r'<acronym title="\2">\1</acronym>',
        ),  # 3+ uppercase acronym
        (re.compile(r'\b([A-Z][A-Z\'\-]+[A-Z])(?=[\s.,\)>])'), r'<span class="caps">\1</span>'),  # 3+ uppercase
        (re.compile(r'\b(\s{0,1})?\.{3}'), r'\1&#8230;'),  # ellipsis
        (re.compile(r'^[\*_-]{3,}$', re.M), r'<hr />'),  # <hr> scene-break
        (re.compile(r'(^|[^-])--([^-]|$)'), r'\1&#8212;\2'),  # em dash
        (re.compile(r'\s-(?:\s|$)'), r' &#8211; '),  # en dash
        (re.compile(r'\b( ?)[([]TM[])]', re.I), r'\1&#8482;'),  # trademark
        (re.compile(r'\b( ?)[([]R[])]', re.I), r'\1&#174;'),  # registered
        (re.compile(r'\b( ?)[([]C[])]', re.I), r'\1&#169;'),  # copyright
    ]

    def __init__(self, restricted=False, lite=False, noimage=False):
        """docstring for __init__"""
        self.restricted = restricted
        self.lite = lite
        self.noimage = noimage
        self.get_sizes = False
        self.fn = {}
        self.urlrefs = {}
        self.shelf = {}
        self.rel = ''
        self.html_type = 'xhtml'

    def textile(self, text, rel=None, head_offset=0, html_type='xhtml'):
        """
        >>> import textile
        >>> textile.textile('some textile')
        '\\t<p>some textile</p>'
        """
        self.html_type = html_type

        # text = str(text)
        text = _normalize_newlines(text)

        if self.restricted:
            text = self.encode_html(text, quotes=False)

        if rel:
            self.rel = f' rel="{rel}"'

        text = self.getRefs(text)
        text = self.block(text, int(head_offset))
        text = self.retrieve(text)
        text = smartyPants(text, 'q')

        return text

    def pba(self, input, element=None):
        '''
        Parse block attributes.

        >>> t = Textile()
        >>> t.pba(r'\3')
        ''
        >>> t.pba(r'\\3', element='td')
        ' colspan="3"'
        >>> t.pba(r'/4', element='td')
        ' rowspan="4"'
        >>> t.pba(r'\\3/4', element='td')
        ' colspan="3" rowspan="4"'

        >>> t.vAlign('^')
        'top'

        >>> t.pba('^', element='td')
        ' style="vertical-align:top;"'

        >>> t.pba('{line-height:18px}')
        ' style="line-height:18px;"'

        >>> t.pba('(foo-bar)')
        ' class="foo-bar"'

        >>> t.pba('(#myid)')
        ' id="myid"'

        >>> t.pba('(foo-bar#myid)')
        ' class="foo-bar" id="myid"'

        >>> t.pba('((((')
        ' style="padding-left:4em;"'

        >>> t.pba(')))')
        ' style="padding-right:3em;"'

        >>> t.pba('[fr]')
        ' lang="fr"'

        '''
        style = []
        aclass = ''
        lang = ''
        colspan = ''
        rowspan = ''
        id = ''

        if not input:
            return ''

        matched = input
        if element == 'td':
            m = re.search(r'\\(\d+)', matched)
            if m:
                colspan = m.group(1)

            m = re.search(r'/(\d+)', matched)
            if m:
                rowspan = m.group(1)

        if element in {'td', 'tr'}:
            m = re.search(rf'({self.vlgn})', matched)
            if m:
                style.append(f'vertical-align:{self.vAlign(m.group(1))};')

        m = re.search(r'\{([^}]*)\}', matched)
        if m:
            style.append(m.group(1).rstrip(';') + ';')
            matched = matched.replace(m.group(0), '')

        m = re.search(r'\[([^\]]+)\]', matched, re.U)
        if m:
            lang = m.group(1)
            matched = matched.replace(m.group(0), '')

        m = re.search(r'\(([^()]+)\)', matched, re.U)
        if m:
            aclass = m.group(1)
            matched = matched.replace(m.group(0), '')

        m = re.search(r'([(]+)', matched)
        if m:
            style.append(f'padding-left:{len(m.group(1))}em;')
            matched = matched.replace(m.group(0), '')

        m = re.search(r'([)]+)', matched)
        if m:
            style.append(f'padding-right:{len(m.group(1))}em;')
            matched = matched.replace(m.group(0), '')

        m = re.search(rf'({self.hlgn})', matched)
        if m:
            style.append(f'text-align:{self.hAlign(m.group(1))};')

        m = re.search(r'^(.*)#(.*)$', aclass)
        if m:
            id = m.group(2)
            aclass = m.group(1)

        if self.restricted:
            if lang:
                return ' lang="%s"'
            else:
                return ''

        result = []
        if style:
            result.append(' style="{}"'.format(''.join(style)))
        if aclass:
            result.append(f' class="{aclass}"')
        if lang:
            result.append(f' lang="{lang}"')
        if id:
            result.append(f' id="{id}"')
        if colspan:
            result.append(f' colspan="{colspan}"')
        if rowspan:
            result.append(f' rowspan="{rowspan}"')
        return ''.join(result)

    def hasRawText(self, text):
        """
        checks whether the text has text not already enclosed by a block tag

        >>> t = Textile()
        >>> t.hasRawText('<p>foo bar biz baz</p>')
        False

        >>> t.hasRawText(' why yes, yes it does')
        True

        """
        r = re.compile(r'<(p|blockquote|div|form|table|ul|ol|pre|h\d)[^>]*?>.*</\1>', re.S).sub('', text.strip()).strip()
        r = re.compile(r'<(hr|br)[^>]*?/>').sub('', r)
        return '' != r

    def table(self, text):
        r"""
        >>> t = Textile()
        >>> t.table('|one|two|three|\n|a|b|c|')
        '\t<table>\n\t\t<tr>\n\t\t\t<td>one</td>\n\t\t\t<td>two</td>\n\t\t\t<td>three</td>\n\t\t</tr>\n\t\t<tr>\n\t\t\t<td>a</td>\n\t\t\t<td>b</td>\n\t\t\t<td>c</td>\n\t\t</tr>\n\t</table>\n\n'
        """
        text = text + '\n\n'
        pattern = re.compile(rf'^(?:table(_?{self.s}{self.a}{self.c})\. ?\n)?^({self.a}{self.c}\.? ?\|.*\|)\n\n', re.S | re.M | re.U)
        return pattern.sub(self.fTable, text)

    def fTable(self, match):
        tatts = self.pba(match.group(1), 'table')
        rows = []
        for row in [x for x in match.group(2).split('\n') if x]:
            rmtch = re.search(rf'^({self.a}{self.c}\. )(.*)', row.lstrip())
            if rmtch:
                ratts = self.pba(rmtch.group(1), 'tr')
                row = rmtch.group(2)
            else:
                ratts = ''

            cells = []
            for cell in row.split('|')[1:-1]:
                ctyp = 'd'
                if re.search(r'^_', cell):
                    ctyp = 'h'
                cmtch = re.search(rf'^(_?{self.s}{self.a}{self.c}\. )(.*)', cell)
                if cmtch:
                    catts = self.pba(cmtch.group(1), 'td')
                    cell = cmtch.group(2)
                else:
                    catts = ''

                cell = self.graf(self.span(cell))
                cells.append(f'\t\t\t<t{ctyp}{catts}>{cell}</t{ctyp}>')
            rows.append('\t\t<tr{}>\n{}\n\t\t</tr>'.format(ratts, '\n'.join(cells)))
            cells = []
            catts = None
        return '\t<table{}>\n{}\n\t</table>\n\n'.format(tatts, '\n'.join(rows))

    def lists(self, text):
        '''
        >>> t = Textile()
        >>> t.lists("* one\\n* two\\n* three")
        '\\t<ul>\\n\\t\\t<li>one</li>\\n\\t\\t<li>two</li>\\n\\t\\t<li>three</li>\\n\\t</ul>'
        '''
        pattern = re.compile(rf'^([#*]+{self.c} .*)$(?![^#*])', re.U | re.M | re.S)
        return pattern.sub(self.fList, text)

    def fList(self, match):
        text = match.group(0).split('\n')
        result = []
        lists = []
        for i, line in enumerate(text):
            try:
                nextline = text[i + 1]
            except IndexError:
                nextline = ''

            m = re.search(rf'^([#*]+)({self.a}{self.c}) (.*)$', line, re.S)
            if m:
                tl, atts, content = m.groups()
                nl = ''
                nm = re.search(r'^([#*]+)\s.*', nextline)
                if nm:
                    nl = nm.group(1)
                if tl not in lists:
                    lists.append(tl)
                    atts = self.pba(atts)
                    line = f'\t<{self.lT(tl)}l{atts}>\n\t\t<li>{self.graf(content)}'
                else:
                    line = '\t\t<li>' + self.graf(content)

                if len(nl) <= len(tl):
                    line = line + '</li>'
                for k in reversed(lists):
                    if len(k) > len(nl):
                        line = line + f'\n\t</{self.lT(k)}l>'
                        if len(k) > 1:
                            line = line + '</li>'
                        lists.remove(k)

            result.append(line)
        return '\n'.join(result)

    def lT(self, input):
        if re.search(r'^#+', input):
            return 'o'
        else:
            return 'u'

    def doPBr(self, in_):
        return re.compile(r'<(p)([^>]*?)>(.*)(</\1>)', re.S).sub(self.doBr, in_)

    def doBr(self, match):
        if self.html_type == 'html':
            content = re.sub(r'(.+)(?:(?<!<br>)|(?<!<br />))\n(?![#*\s|])', '\\1<br>', match.group(3))
        else:
            content = re.sub(r'(.+)(?:(?<!<br>)|(?<!<br />))\n(?![#*\s|])', '\\1<br />', match.group(3))
        return f'<{match.group(1)}{match.group(2)}>{content}{match.group(4)}'

    def block(self, text, head_offset=0):
        """
        >>> t = Textile()
        >>> t.block('h1. foobar baby')
        '\\t<h1>foobar baby</h1>'
        """
        if not self.lite:
            tre = '|'.join(self.btag)
        else:
            tre = '|'.join(self.btag_lite)
        text = text.split('\n\n')

        tag = 'p'
        atts = cite = graf = ext = c1 = ''

        out = []

        anon = False
        for line in text:
            pattern = rf'^({tre})({self.a}{self.c})\.(\.?)(?::(\S+))? (.*)$'
            match = re.search(pattern, line, re.S)
            if match:
                if ext:
                    out.append(out.pop() + c1)

                tag, atts, ext, cite, graf = match.groups()
                h_match = re.search(r'h([1-6])', tag)
                if h_match:
                    (head_level,) = h_match.groups()
                    tag = f'h{max(1, min(int(head_level) + head_offset, 6))}'
                o1, o2, content, c2, c1 = self.fBlock(tag, atts, ext, cite, graf)
                # leave off c1 if this block is extended,
                # we'll close it at the start of the next block

                if ext:
                    line = f'{o1}{o2}{content}{c2}'
                else:
                    line = f'{o1}{o2}{content}{c2}{c1}'

            else:
                anon = True
                if ext or not re.search(r'^\s', line):
                    o1, o2, content, c2, c1 = self.fBlock(tag, atts, ext, cite, line)
                    # skip $o1/$c1 because this is part of a continuing
                    # extended block
                    if tag == 'p' and not self.hasRawText(content):
                        line = content
                    else:
                        line = f'{o2}{content}{c2}'
                else:
                    line = self.graf(line)

            line = self.doPBr(line)
            if self.html_type == 'xhtml':
                line = line.replace('<br>', '<br />')

            if ext and anon:
                out.append(out.pop() + '\n' + line)
            else:
                out.append(line)

            if not ext:
                tag = 'p'
                atts = ''
                cite = ''
                graf = ''

        if ext:
            out.append(out.pop() + c1)
        return '\n\n'.join(out)

    def fBlock(self, tag, atts, ext, cite, content):
        '''
        >>> t = Textile()
        >>> t.fBlock("bq", "", None, "", "Hello BlockQuote")
        ('\\t<blockquote>\\n', '\\t\\t<p>', 'Hello BlockQuote', '</p>', '\\n\\t</blockquote>')

        >>> t.fBlock("bq", "", None, "http://google.com", "Hello BlockQuote")
        ('\\t<blockquote cite="http://google.com">\\n', '\\t\\t<p>', 'Hello BlockQuote', '</p>', '\\n\\t</blockquote>')

        >>> t.fBlock("bc", "", None, "", 'printf "Hello, World";') # doctest: +ELLIPSIS
        ('<pre>', '<code>', ..., '</code>', '</pre>')

        >>> t.fBlock("h1", "", None, "", "foobar")
        ('', '\\t<h1>', 'foobar', '</h1>', '')
        '''
        atts = self.pba(atts)
        o1 = o2 = c2 = c1 = ''

        m = re.search(r'fn(\d+)', tag)
        if m:
            tag = 'p'
            if m.group(1) in self.fn:
                fnid = self.fn[m.group(1)]
            else:
                fnid = m.group(1)
            atts = atts + f' id="fn{fnid}"'
            if atts.find('class=') < 0:
                atts = atts + ' class="footnote"'
            content = (f'<sup>{m.group(1)}</sup>') + content

        if tag == 'bq':
            cite = self.checkRefs(cite)
            if cite:
                cite = f' cite="{cite}"'
            else:
                cite = ''
            o1 = f'\t<blockquote{cite}{atts}>\n'
            o2 = f'\t\t<p{atts}>'
            c2 = '</p>'
            c1 = '\n\t</blockquote>'

        elif tag == 'bc':
            o1 = f'<pre{atts}>'
            o2 = f'<code{atts}>'
            c2 = '</code>'
            c1 = '</pre>'
            content = self.shelve(self.encode_html(content.rstrip('\n') + '\n'))

        elif tag == 'notextile':
            content = self.shelve(content)
            o1 = o2 = ''
            c1 = c2 = ''

        elif tag == 'pre':
            content = self.shelve(self.encode_html(content.rstrip('\n') + '\n'))
            o1 = f'<pre{atts}>'
            o2 = c2 = ''
            c1 = '</pre>'

        else:
            o2 = f'\t<{tag}{atts}>'
            c2 = f'</{tag}>'

        content = self.graf(content)
        return o1, o2, content, c2, c1

    def footnoteRef(self, text):
        '''
        >>> t = Textile()
        >>> t.footnoteRef('foo[1] ') # doctest: +ELLIPSIS
        'foo<sup class="footnote"><a href="#fn...">1</a></sup> '
        '''
        return re.sub(r'\b\[([0-9]+)\](\s)?', self.footnoteID, text)

    def footnoteID(self, match):
        id, t = match.groups()
        if id not in self.fn:
            self.fn[id] = str(uuid.uuid4())
        fnid = self.fn[id]
        if not t:
            t = ''
        return f'<sup class="footnote"><a href="#fn{fnid}">{id}</a></sup>{t}'

    def glyphs(self, text):
        '''
        >>> t = Textile()

        >>> t.glyphs("apostrophe's")
        'apostrophe&#8217;s'

        >>> t.glyphs("back in '88")
        'back in &#8217;88'

        >>> t.glyphs('foo ...')
        'foo &#8230;'

        >>> t.glyphs('--')
        '&#8212;'

        >>> t.glyphs('FooBar[tm]')
        'FooBar&#8482;'

        >>> t.glyphs("<p><cite>Cat's Cradle</cite> by Vonnegut</p>")
        '<p><cite>Cat&#8217;s Cradle</cite> by Vonnegut</p>'

        '''
        # fix: hackish
        text = re.sub(r'"\Z', '" ', text)

        result = []
        for line in re.compile(r'(<.*?>)', re.U).split(text):
            if not re.search(r'<.*>', line):
                rules = []
                if re.search(r'{.+?}', line):
                    rules = self.macro_defaults + self.glyph_defaults
                else:
                    rules = self.glyph_defaults
                for s, r in rules:
                    line = s.sub(r, line)
            result.append(line)
        return ''.join(result)

    def macros_only(self, text):
        # fix: hackish
        text = re.sub(r'"\Z', '" ', text)

        result = []
        for line in re.compile(r'(<.*?>)', re.U).split(text):
            if not re.search(r'<.*>', line):
                rules = []
                if re.search(r'{.+?}', line):
                    rules = self.macro_defaults
                for s, r in rules:
                    line = s.sub(r, line)
            result.append(line)
        return ''.join(result)

    def vAlign(self, input):
        d = {'^': 'top', '-': 'middle', '~': 'bottom'}
        return d.get(input, '')

    def hAlign(self, input):
        d = {'<': 'left', '=': 'center', '>': 'right', '<>': 'justify'}
        return d.get(input, '')

    def getRefs(self, text):
        """
        what is this for?
        """
        pattern = re.compile(r'(?:(?<=^)|(?<=\s))\[(.+)\]((?:http(?:s?)://|/)\S+)(?=\s|$)', re.U)
        text = pattern.sub(self.refs, text)
        return text

    def refs(self, match):
        flag, url = match.groups()
        self.urlrefs[flag] = url
        return ''

    def checkRefs(self, url):
        return self.urlrefs.get(url, url)

    def isRelURL(self, url):
        '''
        Identify relative urls.

        >>> t = Textile()
        >>> t.isRelURL("http://www.google.com/")
        False
        >>> t.isRelURL("/foo")
        True

        '''
        scheme, netloc = urlparse(url)[0:2]
        return not scheme and not netloc

    def relURL(self, url):
        scheme = urlparse(url)[0]
        if self.restricted and scheme and scheme not in self.url_schemes:
            return '#'
        return url

    def shelve(self, text):
        id = str(uuid.uuid4()) + 'c'
        self.shelf[id] = text
        return id

    def retrieve(self, text):
        '''
        >>> t = Textile()
        >>> id = t.shelve("foobar")
        >>> t.retrieve(id)
        'foobar'
        '''
        while True:
            old = text
            for k, v in self.shelf.items():
                text = text.replace(k, v)
            if text == old:
                break
        return text

    def encode_html(self, text, quotes=True):
        a = (('&', '&#38;'), ('<', '&#60;'), ('>', '&#62;'))

        if quotes:
            a = a + (("'", '&#39;'), ('"', '&#34;'))

        for k, v in a:
            text = text.replace(k, v)
        return text

    def graf(self, text):
        if not self.lite:
            text = self.noTextile(text)
            text = self.code(text)

        text = self.links(text)

        if not self.noimage:
            text = self.image(text)

        if not self.lite:
            text = self.lists(text)
            text = self.table(text)

        text = self.span(text)
        text = self.footnoteRef(text)
        text = self.glyphs(text)

        return text.rstrip('\n')

    def links(self, text):
        '''
        >>> t = Textile()
        >>> t.links('fooobar "Google":http://google.com/foobar/ and hello world "flickr":http://flickr.com/photos/jsamsa/ ') # doctest: +ELLIPSIS
        'fooobar ... and hello world ...'
        '''

        text = self.macros_only(text)
        punct = '!"#$%&\'*+,-./:;=?@\\^_`|~'

        pattern = rf'''
            (?P<pre>    [\s\[{{(]|[{re.escape(punct)}]   )?
            "                          # start
            (?P<atts>   {self.c}       )
            (?P<text>   [^"]+?   )
            \s?
            (?:   \(([^)]+?)\)(?=")   )?     # $title
            ":
            (?P<url>    (?:ftp|https?)? (?: :// )? [-A-Za-z0-9+&@#/?=~_()|!:,.;]*[-A-Za-z0-9+&@#/=~_()|]   )
            (?P<post>   [^\w\/;]*?   )
            (?=<|\s|$)
        '''

        text = re.compile(pattern, re.X).sub(self.fLink, text)

        return text

    def fLink(self, match):
        pre, atts, text, title, url, post = match.groups()

        if pre is None:
            pre = ''

        # assume ) at the end of the url is not actually part of the url
        # unless the url also contains a (
        if url.endswith(')') and not url.find('(') > -1:
            post = url[-1] + post
            url = url[:-1]

        url = self.checkRefs(url)

        atts = self.pba(atts)
        if title:
            atts = atts + f' title="{self.encode_html(title)}"'

        if not self.noimage:
            text = self.image(text)

        text = self.span(text)
        text = self.glyphs(text)

        url = self.relURL(url)
        out = f'<a href="{self.encode_html(url)}"{atts}{self.rel}>{text}</a>'
        out = self.shelve(out)
        return ''.join([pre, out, post])

    def span(self, text):
        '''
        >>> t = Textile()
        >>> t.span(r"hello %(bob)span *strong* and **bold**% goodbye")
        'hello <span class="bob">span <strong>strong</strong> and <b>bold</b></span> goodbye'
        '''
        qtags = (r'\*\*', r'\*', r'\?\?', r'\-', r'__', r'_', r'%', r'\+', r'~', r'\^')
        pnct = ".,\"'?!;:"

        for qtag in qtags:
            pattern = re.compile(
                rf'''
                (?:^|(?<=[\s>{pnct}\(])|\[|([\]}}]))
                ({qtag})(?!{qtag})
                ({self.c})
                (?::(\S+))?
                ([^\s{qtag}]+|\S[^{qtag}\n]*[^\s{qtag}\n])
                ([{pnct}]*)
                {qtag}
                (?:$|([\]}}])|(?={self.pnct}{{1,2}}|\s))
            ''',
                re.X,
            )
            text = pattern.sub(self.fSpan, text)
        return text

    def fSpan(self, match):
        _, tag, atts, cite, content, end, _ = match.groups()

        qtags = {
            '*': 'strong',
            '**': 'b',
            '??': 'cite',
            '_': 'em',
            '__': 'i',
            '-': 'del',
            '%': 'span',
            '+': 'ins',
            '~': 'sub',
            '^': 'sup',
        }
        tag = qtags[tag]
        atts = self.pba(atts)
        if cite:
            atts = atts + f'cite="{cite}"'

        content = self.span(content)

        out = f'<{tag}{atts}>{content}{end}</{tag}>'
        return out

    def image(self, text):
        '''
        >>> t = Textile()
        >>> t.image('!/imgs/myphoto.jpg!:http://jsamsa.com')
        '<a href="http://jsamsa.com"><img src="/imgs/myphoto.jpg" alt="" /></a>'
        '''
        pattern = re.compile(
            rf'''
            (?:[\[{{])?          # pre
            \!                 # opening !
            ({self.c})               # optional style,class atts
            (?:\. )?           # optional dot-space
            ([^\s(!]+)         # presume this is the src
            \s?                # optional space
            (?:\(([^\)]+)\))?  # optional title
            \!                 # closing
            (?::(\S+))?        # optional href
            (?:[\]}}]|(?=\s|$)) # lookahead: space or end of string
        ''',
            re.U | re.X,
        )
        return pattern.sub(self.fImage, text)

    def fImage(self, match):
        # (None, '', '/imgs/myphoto.jpg', None, None)
        atts, url, title, href = match.groups()
        atts = self.pba(atts)

        if title:
            atts = atts + f' title="{title}" alt="{title}"'
        else:
            atts = atts + ' alt=""'

        if not self.isRelURL(url) and self.get_sizes:
            size = getimagesize(url)
            if size:
                atts += f' {size}'

        if href:
            href = self.checkRefs(href)

        url = self.checkRefs(url)
        url = self.relURL(url)

        out = []
        if href:
            out.append(f'<a href="{href}" class="img">')
        if self.html_type == 'html':
            out.append(f'<img src="{url}"{atts}>')
        else:
            out.append(f'<img src="{url}"{atts} />')
        if href:
            out.append('</a>')

        return ''.join(out)

    def code(self, text):
        text = self.doSpecial(text, '<code>', '</code>', self.fCode)
        text = self.doSpecial(text, '@', '@', self.fCode)
        text = self.doSpecial(text, '<pre>', '</pre>', self.fPre)
        return text

    def fCode(self, match):
        before, text, after = match.groups()
        if after is None:
            after = ''
        # text needs to be escaped
        if not self.restricted:
            text = self.encode_html(text)
        return ''.join([before, self.shelve(f'<code>{text}</code>'), after])

    def fPre(self, match):
        before, text, after = match.groups()
        if after is None:
            after = ''
        # text needs to be escapedd
        if not self.restricted:
            text = self.encode_html(text)
        return ''.join([before, '<pre>', self.shelve(text), '</pre>', after])

    def doSpecial(self, text, start, end, method=None):
        if method is None:
            method = self.fSpecial
        pattern = re.compile(rf'(^|\s|[\[({{>]){re.escape(start)}(.*?){re.escape(end)}(\s|$|[\])}}])?', re.M | re.S)
        return pattern.sub(method, text)

    def fSpecial(self, match):
        """
        special blocks like notextile or code
        """
        before, text, after = match.groups()
        if after is None:
            after = ''
        return ''.join([before, self.shelve(self.encode_html(text)), after])

    def noTextile(self, text):
        text = self.doSpecial(text, '<notextile>', '</notextile>', self.fTextile)
        return self.doSpecial(text, '==', '==', self.fTextile)

    def fTextile(self, match):
        before, notextile, after = match.groups()
        if after is None:
            after = ''
        return ''.join([before, self.shelve(notextile), after])


def textile(text, head_offset=0, html_type='xhtml', encoding=None, output=None):
    """
    this function takes additional parameters:
    head_offset - offset to apply to heading levels (default: 0)
    html_type - 'xhtml' or 'html' style tags (default: 'xhtml')
    """
    return Textile().textile(text, head_offset=head_offset, html_type=html_type)


def textile_restricted(text, lite=True, noimage=True, html_type='xhtml'):
    """
    Restricted version of Textile designed for weblog comments and other
    untrusted input.

    Raw HTML is escaped.
    Style attributes are disabled.
    rel='nofollow' is added to external links.

    When lite=True is set (the default):
    Block tags are restricted to p, bq, and bc.
    Lists and tables are disabled.

    When noimage=True is set (the default):
    Image tags are disabled.

    """
    return Textile(restricted=True, lite=lite, noimage=noimage).textile(text, rel='nofollow', html_type=html_type)



# Form implementation generated from reading ui file '/home/kovid/work/calibre/src/calibre/gui2/dialogs/jobs.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_JobsDialog(object):
    def setupUi(self, JobsDialog):
        JobsDialog.setObjectName("JobsDialog")
        JobsDialog.resize(633, 542)
        icon = QtGui.QIcon.ic("jobs.png")
        JobsDialog.setWindowIcon(icon)
        self.gridLayout = QtWidgets.QGridLayout(JobsDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.search = SearchBox2(parent=JobsDialog)
        self.search.setObjectName("search")
        self.horizontalLayout.addWidget(self.search)
        self.search_button = QtWidgets.QToolButton(parent=JobsDialog)
        icon1 = QtGui.QIcon.ic("search.png")
        self.search_button.setIcon(icon1)
        self.search_button.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self.search_button.setObjectName("search_button")
        self.horizontalLayout.addWidget(self.search_button)
        self.gridLayout.addLayout(self.horizontalLayout, 0, 0, 1, 2)
        self.jobs_view = QtWidgets.QTableView(parent=JobsDialog)
        self.jobs_view.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.NoContextMenu)
        self.jobs_view.setEditTriggers(QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers)
        self.jobs_view.setAlternatingRowColors(True)
        self.jobs_view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
        self.jobs_view.setIconSize(QtCore.QSize(32, 32))
        self.jobs_view.setObjectName("jobs_view")
        self.gridLayout.addWidget(self.jobs_view, 1, 0, 1, 2)
        self.kill_button = QtWidgets.QPushButton(parent=JobsDialog)
        self.kill_button.setObjectName("kill_button")
        self.gridLayout.addWidget(self.kill_button, 2, 0, 1, 1)
        self.hide_button = QtWidgets.QPushButton(parent=JobsDialog)
        self.hide_button.setObjectName("hide_button")
        self.gridLayout.addWidget(self.hide_button, 2, 1, 1, 1)
        self.details_button = QtWidgets.QPushButton(parent=JobsDialog)
        self.details_button.setObjectName("details_button")
        self.gridLayout.addWidget(self.details_button, 3, 0, 1, 1)
        self.show_button = QtWidgets.QPushButton(parent=JobsDialog)
        self.show_button.setObjectName("show_button")
        self.gridLayout.addWidget(self.show_button, 3, 1, 1, 1)
        self.stop_all_jobs_button = QtWidgets.QPushButton(parent=JobsDialog)
        self.stop_all_jobs_button.setObjectName("stop_all_jobs_button")
        self.gridLayout.addWidget(self.stop_all_jobs_button, 4, 0, 1, 1)
        self.hide_all_button = QtWidgets.QPushButton(parent=JobsDialog)
        self.hide_all_button.setObjectName("hide_all_button")
        self.gridLayout.addWidget(self.hide_all_button, 4, 1, 1, 1)
        self.auto_close_calibre = QtWidgets.QCheckBox(parent=JobsDialog)
        icon2 = QtGui.QIcon.ic("dialog_warning.png")
        self.auto_close_calibre.setIcon(icon2)
        self.auto_close_calibre.setObjectName("auto_close_calibre")
        self.gridLayout.addWidget(self.auto_close_calibre, 5, 0, 1, 2)

        self.retranslateUi(JobsDialog)
        QtCore.QMetaObject.connectSlotsByName(JobsDialog)

    def retranslateUi(self, JobsDialog):

        JobsDialog.setWindowTitle(_("Active jobs"))
        self.search_button.setToolTip(_("Find next match"))
        self.search_button.setText(_("&Search"))
        self.kill_button.setText(_("&Stop selected jobs"))
        self.hide_button.setText(_("&Hide selected jobs"))
        self.details_button.setText(_("Show job &details"))
        self.show_button.setText(_("Show &all jobs"))
        self.stop_all_jobs_button.setText(_("Stop &all non device jobs"))
        self.hide_all_button.setText(_("&Hide all jobs"))
        self.auto_close_calibre.setText(_("&Automatically close calibre after all jobs are finished"))
        self.auto_close_calibre.setToolTip(_("Once all currently running jobs are completed, calibre will automatically quit. Before quitting, a popup will be shown for thirty seconds allowing you to cancel the automatic quit."))
from calibre.gui2.search_box import SearchBox2



#!/usr/bin/env python
# License: GPLv3 Copyright: 2011, Kovid Goyal <kovid@kovidgoyal.net>

import sys
from contextlib import suppress

from qt.core import (
    QAction,
    QApplication,
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QIcon,
    QKeySequence,
    QLabel,
    QPainter,
    QPlainTextEdit,
    QSize,
    QSizePolicy,
    Qt,
    QTextBrowser,
    QTextDocument,
    QTimer,
    QVBoxLayout,
    QWidget,
    pyqtSignal,
)

from calibre.constants import __version__, isfrozen
from calibre.gui2 import gprefs
from calibre.utils.localization import _, ngettext


class Icon(QWidget):
    def __init__(self, parent=None, size=None):
        QWidget.__init__(self, parent)
        self.pixmap = None
        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.icon_size = size or 64

    def set_icon(self, qicon):
        self.pixmap = qicon.pixmap(self.icon_size, self.icon_size)
        self.update()

    def sizeHint(self):
        return QSize(self.icon_size, self.icon_size)

    def paintEvent(self, a0):
        if self.pixmap is not None:
            x = (self.width() - self.icon_size) // 2
            y = (self.height() - self.icon_size) // 2
            p = QPainter(self)
            p.drawPixmap(x, y, self.icon_size, self.icon_size, self.pixmap)


def might_be_rich_text(text: str) -> bool:
    import re

    if not text:
        return False

    # Qt looks at the text before the first line break
    first_line = text.split('\n', 1)[0]

    # 1. Check for common HTML entity references (e.g., &amp;, &#123;)
    if '&' in first_line and re.search(r'&#?[a-zA-Z0-9]+;', first_line):
        return True

    # 2. Check for an HTML tag sequence: <tag...>, </tag>, or <tag/>
    # Qt looks for '<' followed by an alphanumeric character, '/' or '!'
    if '<' in first_line and re.search(r'<[a-zA-Z!/]', first_line):
        return True

    return False


class MessageBox(QDialog):  # {{{
    ERROR = 0
    WARNING = 1
    INFO = 2
    QUESTION = 3

    resize_needed = pyqtSignal()

    def setup_ui(self):
        self.setObjectName('Dialog')
        self.resize(497, 235)
        self.gridLayout = l = QGridLayout(self)
        l.setObjectName('gridLayout')
        self.icon_widget = Icon(self)
        l.addWidget(self.icon_widget)
        self.msg = la = QLabel(self)
        la.setWordWrap(True), la.setMinimumWidth(400)
        la.setOpenExternalLinks(True)
        la.setObjectName('msg')
        l.addWidget(la, 0, 1, 1, 1)
        self.det_msg = dm = QTextBrowser(self)
        dm.setReadOnly(True)
        dm.setObjectName('det_msg')
        l.addWidget(dm, 1, 0, 1, 2)
        self.bb = bb = QDialogButtonBox(self)
        bb.setStandardButtons(QDialogButtonBox.StandardButton.Ok)
        bb.setObjectName('bb')
        bb.accepted.connect(self.accept)
        bb.rejected.connect(self.reject)
        l.addWidget(bb, 3, 0, 1, 2)
        self.toggle_checkbox = tc = QCheckBox(self)
        tc.setObjectName('toggle_checkbox')
        l.addWidget(tc, 2, 0, 1, 2)

    def __init__(
        self,
        type_,
        title,
        msg,
        det_msg='',
        q_icon=None,
        show_copy_button=True,
        parent=None,
        default_yes=True,
        yes_text=None,
        no_text=None,
        yes_icon=None,
        no_icon=None,
        add_abort_button=False,
        only_copy_details=False,
    ):
        super().__init__(parent)
        self.only_copy_details = only_copy_details
        self.aborted = False
        if q_icon is None:
            icon = {
                self.ERROR: 'error',
                self.WARNING: 'warning',
                self.INFO: 'information',
                self.QUESTION: 'question',
            }[type_]
            icon = f'dialog_{icon}.png'
            self.icon = QIcon.ic(icon)
        else:
            self.icon = q_icon if isinstance(q_icon, QIcon) else QIcon.ic(q_icon)
        self.setup_ui()

        self.setWindowTitle(title)
        self.setWindowIcon(self.icon)
        self.icon_widget.set_icon(self.icon)
        self.msg.setText(msg)
        if det_msg and might_be_rich_text(det_msg):
            self.det_msg.setHtml(det_msg)
        else:
            self.det_msg.setPlainText(det_msg)
        self.det_msg.setVisible(False)
        self.toggle_checkbox.setVisible(False)

        if show_copy_button:
            self.ctc_button = self.bb.addButton(_('&Copy to clipboard'), QDialogButtonBox.ButtonRole.ActionRole)
            assert self.ctc_button is not None
            self.ctc_button.clicked.connect(self.copy_to_clipboard)

        self.show_det_msg = _('Show &details')
        self.hide_det_msg = _('Hide &details')
        self.det_msg_toggle = self.bb.addButton(self.show_det_msg, QDialogButtonBox.ButtonRole.ActionRole)
        det_msg_toggle = self.det_msg_toggle
        assert det_msg_toggle is not None
        det_msg_toggle.clicked.connect(self.toggle_det_msg)
        det_msg_toggle.setToolTip(_('Show detailed information about this error'))

        self.copy_action = QAction(self)
        self.addAction(self.copy_action)
        self.copy_action.setShortcuts(QKeySequence.StandardKey.Copy)
        self.copy_action.triggered.connect(self.copy_to_clipboard)

        self.is_question = type_ == self.QUESTION
        if self.is_question:
            self.bb.setStandardButtons(QDialogButtonBox.StandardButton.Yes | QDialogButtonBox.StandardButton.No)
            default_btn = self.bb.button(QDialogButtonBox.StandardButton.Yes if default_yes else QDialogButtonBox.StandardButton.No)
            assert default_btn is not None
            default_btn.setDefault(True)
            self.default_yes = default_yes
            if yes_text is not None:
                yes_btn = self.bb.button(QDialogButtonBox.StandardButton.Yes)
                assert yes_btn is not None
                yes_btn.setText(yes_text)
            if no_text is not None:
                no_btn = self.bb.button(QDialogButtonBox.StandardButton.No)
                assert no_btn is not None
                no_btn.setText(no_text)
            if yes_icon is not None:
                yes_icon_btn = self.bb.button(QDialogButtonBox.StandardButton.Yes)
                assert yes_icon_btn is not None
                yes_icon_btn.setIcon(yes_icon if isinstance(yes_icon, QIcon) else QIcon.ic(yes_icon))
            if no_icon is not None:
                no_icon_btn = self.bb.button(QDialogButtonBox.StandardButton.No)
                assert no_icon_btn is not None
                no_icon_btn.setIcon(no_icon if isinstance(no_icon, QIcon) else QIcon.ic(no_icon))
        else:
            ok_btn = self.bb.button(QDialogButtonBox.StandardButton.Ok)
            assert ok_btn is not None
            ok_btn.setDefault(True)

        if add_abort_button:
            abort_btn = self.bb.addButton(QDialogButtonBox.StandardButton.Abort)
            assert abort_btn is not None
            abort_btn.clicked.connect(self.on_abort)

        if not det_msg:
            det_msg_toggle.setVisible(False)

        self.resize_needed.connect(self.do_resize)
        self.do_resize()

    def on_abort(self):
        self.aborted = True

    def toggle_det_msg(self, *args):
        vis = self.det_msg.isVisible()
        self.det_msg.setVisible(not vis)
        det_msg_toggle = self.det_msg_toggle
        assert det_msg_toggle is not None
        det_msg_toggle.setText(self.show_det_msg if vis else self.hide_det_msg)
        self.resize_needed.emit()

    def do_resize(self):
        sz = self.sizeHint()
        sz.setWidth(max(min(sz.width(), 500), self.bb.sizeHint().width() + 100))
        sz.setHeight(min(sz.height(), 500))
        self.setMaximumSize(sz)
        self.resize(self.sizeHint())

    def copy_to_clipboard(self, *args):
        text = self.det_msg.toPlainText()
        if not self.only_copy_details:
            text = f'calibre, version {__version__}\n{self.windowTitle()}: {self.msg.text()}\n\n{text}'
        clipboard = QApplication.clipboard()
        assert clipboard is not None
        clipboard.setText(text)
        if hasattr(self, 'ctc_button'):
            ctc_button = self.ctc_button
            assert ctc_button is not None
            ctc_button.setText(_('Copied'))

    def showEvent(self, a0):
        ret = QDialog.showEvent(self, a0)
        if self.is_question:
            with suppress(Exception):
                focus_btn = self.bb.button(QDialogButtonBox.StandardButton.Yes if self.default_yes else QDialogButtonBox.StandardButton.No)
                assert focus_btn is not None
                focus_btn.setFocus(Qt.FocusReason.OtherFocusReason)
        else:
            ok_btn = self.bb.button(QDialogButtonBox.StandardButton.Ok)
            assert ok_btn is not None
            ok_btn.setFocus(Qt.FocusReason.OtherFocusReason)
        return ret

    def set_details(self, msg):
        if not msg:
            msg = ''
        if might_be_rich_text(msg):
            self.det_msg.setHtml(msg)
        else:
            self.det_msg.setPlainText(msg)
        det_msg_toggle = self.det_msg_toggle
        assert det_msg_toggle is not None
        det_msg_toggle.setText(self.show_det_msg)
        det_msg_toggle.setVisible(bool(msg))
        self.det_msg.setVisible(False)
        self.resize_needed.emit()


# }}}


class ViewLog(QDialog):  # {{{
    def __init__(self, title, html, parent=None, unique_name=None):
        QDialog.__init__(self, parent)
        self.l = l = QVBoxLayout()
        self.setLayout(l)

        self.tb = QTextBrowser(self)
        self.tb.setHtml(f'<pre style="font-family: monospace">{html}</pre>')
        l.addWidget(self.tb)

        self.bb = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok)
        self.bb.accepted.connect(self.accept)
        self.bb.rejected.connect(self.reject)
        self.copy_button = self.bb.addButton(_('Copy to clipboard'), QDialogButtonBox.ButtonRole.ActionRole)
        assert self.copy_button is not None
        self.copy_button.setIcon(QIcon.ic('edit-copy.png'))
        self.copy_button.clicked.connect(self.copy_to_clipboard)
        l.addWidget(self.bb)

        self.unique_name = unique_name or 'view-log-dialog'
        self.finished.connect(self.dialog_closing)
        self.restore_geometry(gprefs, self.unique_name)

        self.setModal(False)
        self.setWindowTitle(title)
        self.setWindowIcon(QIcon.ic('debug.png'))
        self.show()

    def sizeHint(self):
        return QSize(700, 500)

    def copy_to_clipboard(self):
        txt = self.tb.toPlainText()
        clipboard = QApplication.clipboard()
        assert clipboard is not None
        clipboard.setText(txt)

    def dialog_closing(self, result):
        self.save_geometry(gprefs, self.unique_name)


# }}}

_proceed_memory = []


class ProceedNotification(MessageBox):  # {{{
    """
    WARNING: This class is deprecated. DO not use it as some users have
    reported crashes when closing the dialog box generated by this class.
    Instead use: gui.proceed_question(...) The arguments are the same as for
    this class.
    """

    def __init__(
        self,
        callback,
        payload,
        html_log,
        log_viewer_title,
        title,
        msg,
        det_msg='',
        show_copy_button=False,
        parent=None,
        cancel_callback=None,
        log_is_file=False,
    ):
        """
        A non modal popup that notifies the user that a background task has
        been completed.

        :param callback: A callable that is called with payload if the user
        asks to proceed. Note that this is always called in the GUI thread.
        :param cancel_callback: A callable that is called with the payload if
        the users asks not to proceed.
        :param payload: Arbitrary object, passed to callback
        :param html_log: An HTML or plain text log
        :param log_viewer_title: The title for the log viewer window
        :param title: The title for this popup
        :param msg: The msg to display
        :param det_msg: Detailed message
        :param log_is_file: If True the html_log parameter is interpreted as
        the path to a file on disk containing the log encoded with utf-8
        """
        MessageBox.__init__(self, MessageBox.QUESTION, title, msg, det_msg=det_msg, show_copy_button=show_copy_button, parent=parent)
        self.payload = payload
        self.html_log = html_log
        self.log_is_file = log_is_file
        self.log_viewer_title = log_viewer_title

        self.vlb = self.bb.addButton(_('&View log'), QDialogButtonBox.ButtonRole.ActionRole)
        assert self.vlb is not None
        self.vlb.setIcon(QIcon.ic('debug.png'))
        self.vlb.clicked.connect(self.show_log)
        det_msg_toggle = self.det_msg_toggle
        assert det_msg_toggle is not None
        det_msg_toggle.setVisible(bool(det_msg))
        self.setModal(False)
        self.callback, self.cancel_callback = callback, cancel_callback
        _proceed_memory.append(self)

    def show_log(self):
        log = self.html_log
        if self.log_is_file:
            with open(log, 'rb') as f:
                log = f.read().decode('utf-8')
        self.log_viewer = ViewLog(self.log_viewer_title, log, parent=self)

    def do_proceed(self, result):
        from calibre.gui2.ui import get_gui

        func = self.callback if result == QDialog.DialogCode.Accepted else self.cancel_callback
        gui = get_gui(fail_if_absent=True)
        gui.proceed_requested.emit(func, self.payload)
        # Ensure this notification is garbage collected
        vlb = self.vlb
        assert vlb is not None
        vlb.clicked.disconnect()
        self.callback = self.cancel_callback = self.payload = None
        self.setParent(None)
        _proceed_memory.remove(self)

    def done(self, a0):
        self.do_proceed(a0)
        return MessageBox.done(self, a0)


# }}}


class ErrorNotification(MessageBox):  # {{{
    def __init__(self, html_log, log_viewer_title, title, msg, det_msg='', show_copy_button=False, parent=None):
        """
        A non modal popup that notifies the user that a background task has
        errored.

        :param html_log: An HTML or plain text log
        :param log_viewer_title: The title for the log viewer window
        :param title: The title for this popup
        :param msg: The msg to display
        :param det_msg: Detailed message
        """
        MessageBox.__init__(self, MessageBox.ERROR, title, msg, det_msg=det_msg, show_copy_button=show_copy_button, parent=parent)
        self.html_log = html_log
        self.log_viewer_title = log_viewer_title
        self.finished.connect(self.do_close, type=Qt.ConnectionType.QueuedConnection)

        self.vlb = self.bb.addButton(_('&View log'), QDialogButtonBox.ButtonRole.ActionRole)
        assert self.vlb is not None
        self.vlb.setIcon(QIcon.ic('debug.png'))
        self.vlb.clicked.connect(self.show_log)
        det_msg_toggle = self.det_msg_toggle
        assert det_msg_toggle is not None
        det_msg_toggle.setVisible(bool(det_msg))
        self.setModal(False)
        _proceed_memory.append(self)

    def show_log(self):
        self.log_viewer = ViewLog(self.log_viewer_title, self.html_log, parent=self)

    def do_close(self, result):
        # Ensure this notification is garbage collected
        self.setParent(None)
        self.finished.disconnect()
        vlb = self.vlb
        assert vlb is not None
        vlb.clicked.disconnect()
        _proceed_memory.remove(self)


# }}}


class AutoCloseNotification(MessageBox):  # {{{
    def __init__(self, timeout=30, parent=None):
        super().__init__(
            MessageBox.QUESTION,
            _('calibre is about to quit'),
            _('calibre is about to quit. Proceed or abort?'),
            parent=parent,
            show_copy_button=False,
            q_icon='dialog_warning.png',
            default_yes=True,
            yes_text=None,
            no_text=_('Abort'),
        )
        self.timeout = timeout
        self.timer = QTimer(self)
        self.timer.setInterval(self.timeout * 1000)
        self.timer.timeout.connect(self.accept)
        self.timer.setSingleShot(True)
        self.updater = QTimer(self)
        self.updater.setInterval(1000)
        self.updater.timeout.connect(self.update_texts)
        self.updater.setSingleShot(False)
        self.timer.start()
        self.updater.start()
        self.update_texts()

    def reject(self):
        self.timer.stop()
        self.updater.stop()
        return super().reject()

    def accept(self):
        self.timer.stop()
        self.updater.stop()
        return super().accept()

    def update_texts(self):
        time = self.timer.remainingTime() // 1000
        yes_btn = self.bb.button(QDialogButtonBox.StandardButton.Yes)
        assert yes_btn is not None
        yes_btn.setText(_('Close now ({}s)').format(time))


# }}}


class JobError(QDialog):  # {{{
    WIDTH = 600
    do_pop = pyqtSignal()

    def __init__(self, parent):
        QDialog.__init__(self, parent)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose, False)
        self.queue = []
        self.do_pop.connect(self.pop, type=Qt.ConnectionType.QueuedConnection)

        self._layout = l = QGridLayout()
        self.setLayout(l)
        self.icon = QIcon.ic('dialog_error.png')
        self.setWindowIcon(self.icon)
        self.icon_widget = Icon(self)
        self.icon_widget.set_icon(self.icon)
        self.msg_label = QLabel('<p>&nbsp;')
        self.msg_label.setStyleSheet('QLabel { margin-top: 1ex; }')
        self.msg_label.setWordWrap(True)
        self.msg_label.setTextFormat(Qt.TextFormat.RichText)
        self.det_msg = QPlainTextEdit(self)
        self.det_msg.setVisible(False)

        self.bb = QDialogButtonBox(QDialogButtonBox.StandardButton.Close, parent=self)
        self.bb.accepted.connect(self.accept)
        self.bb.rejected.connect(self.reject)
        self.ctc_button = self.bb.addButton(_('&Copy to clipboard'), QDialogButtonBox.ButtonRole.ActionRole)
        assert self.ctc_button is not None
        self.ctc_button.clicked.connect(self.copy_to_clipboard)
        self.retry_button = self.bb.addButton(_('&Retry'), QDialogButtonBox.ButtonRole.ActionRole)
        assert self.retry_button is not None
        self.retry_button.clicked.connect(self.retry)
        self.retry_func = None
        self.show_det_msg = _('Show &details')
        self.hide_det_msg = _('Hide &details')
        self.det_msg_toggle = self.bb.addButton(self.show_det_msg, QDialogButtonBox.ButtonRole.ActionRole)
        det_msg_toggle = self.det_msg_toggle
        assert det_msg_toggle is not None
        det_msg_toggle.clicked.connect(self.toggle_det_msg)
        det_msg_toggle.setToolTip(_('Show detailed information about this error'))
        self.suppress = QCheckBox(self)

        l.addWidget(self.icon_widget, 0, 0, 1, 1)
        l.addWidget(self.msg_label, 0, 1, 1, 1)
        l.addWidget(self.det_msg, 1, 0, 1, 2)
        l.addWidget(self.suppress, 2, 0, 1, 2, Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignBottom)
        l.addWidget(self.bb, 3, 0, 1, 2, Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignBottom)
        l.setColumnStretch(1, 100)

        self.setModal(False)
        self.suppress.setVisible(False)
        self.do_resize()

    def retry(self):
        if self.retry_func is not None:
            self.accept()
            self.retry_func()

    def update_suppress_state(self):
        self.suppress.setText(ngettext('Hide the remaining error message', 'Hide the {} remaining error messages', len(self.queue)).format(len(self.queue)))
        self.suppress.setVisible(len(self.queue) > 3)
        self.do_resize()

    def copy_to_clipboard(self, *args):
        d = QTextDocument()
        d.setHtml(self.msg_label.text())
        clipboard = QApplication.clipboard()
        assert clipboard is not None
        clipboard.setText(
            f'calibre, version {__version__} ({sys.platform}, embedded-python: {isfrozen})\n'
            f'{self.windowTitle()!s}: {d.toPlainText()!s}\n\n{self.det_msg.toPlainText()!s}'
        )
        if hasattr(self, 'ctc_button'):
            ctc_button = self.ctc_button
            assert ctc_button is not None
            ctc_button.setText(_('Copied'))

    def toggle_det_msg(self, *args):
        det_msg_toggle = self.det_msg_toggle
        assert det_msg_toggle is not None
        vis = str(det_msg_toggle.text()) == self.hide_det_msg
        det_msg_toggle.setText(self.show_det_msg if vis else self.hide_det_msg)
        self.det_msg.setVisible(not vis)
        self.do_resize()

    def do_resize(self):
        h = self.sizeHint().height()
        self.setMinimumHeight(0)  # Needed as this gets set if det_msg is shown
        # Needed otherwise re-showing the box after showing det_msg causes the box
        # to not reduce in height
        self.setMaximumHeight(h)
        self.resize(QSize(self.WIDTH, h))

    def showEvent(self, a0):
        ret = QDialog.showEvent(self, a0)
        close_btn = self.bb.button(QDialogButtonBox.StandardButton.Close)
        assert close_btn is not None
        close_btn.setFocus(Qt.FocusReason.OtherFocusReason)
        return ret

    def show_error(self, title, msg, det_msg='', retry_func=None):
        self.queue.append((title, msg, det_msg, retry_func))
        self.update_suppress_state()
        self.pop()

    def pop(self):
        if not self.queue or self.isVisible():
            return
        title, msg, det_msg, retry_func = self.queue.pop(0)
        self.setWindowTitle(title)
        self.msg_label.setText(msg)
        self.det_msg.setPlainText(det_msg)
        self.det_msg.setVisible(False)
        det_msg_toggle = self.det_msg_toggle
        assert det_msg_toggle is not None
        det_msg_toggle.setText(self.show_det_msg)
        det_msg_toggle.setVisible(True)
        self.suppress.setChecked(False)
        self.update_suppress_state()
        if not det_msg:
            det_msg_toggle.setVisible(False)
        retry_button = self.retry_button
        assert retry_button is not None
        retry_button.setVisible(retry_func is not None)
        self.retry_func = retry_func
        self.do_resize()
        self.show()

    def done(self, a0):
        if self.suppress.isChecked():
            self.queue = []
        QDialog.done(self, a0)
        self.do_pop.emit()


# }}}

if __name__ == '__main__':
    from qt.core import QMainWindow, QTimer

    from calibre import prepare_string_for_xml
    from calibre.gui2 import Application, question_dialog

    app = Application([])
    merged = {'Kovid Goyal': ['Waterloo', 'Doomed'], 'Someone Else': ['Some other book ' * 1000]}
    lines = []
    for author in sorted(merged):
        lines.append(f'<b><i>{prepare_string_for_xml(author)}</i></b><ol style="margin-top: 0">')
        for title in sorted(merged[author]):
            lines.append(f'<li>{prepare_string_for_xml(title)}</li>')
        lines.append('</ol>')
    w = QMainWindow()
    w.show()

    def doit():
        print(question_dialog(w, 'title', 'msg <a href="http://google.com">goog</a> ', det_msg='\n'.join(lines), show_copy_button=True))
        w.close()

    QTimer.singleShot(100, doit)
    app.exec()



#!/usr/bin/env python
# License: GPLv3 Copyright: 2026, Kovid Goyal <kovid at kovidgoyal.net>

import random
from collections.abc import Iterable
from urllib.parse import urlparse

from camoufox.async_api import AsyncCamoufox  # type: ignore

from calibre.constants import ismacos, iswindows

wikipedia_topics = (
    'Lists_of_deaths_by_year',
    'United_States',
    'India',
    'China',
    'United_Kingdom',
    'President',
    'Red',
    'Green',
    'Cyan',
    'Magenta',
)
subreddits = (
    'funny',
    'AskReddit',
    'gaming',
    'pics',
    'worldnews',
    'todayilearned',
    'Music',
    'movies',
    'science',
    'memes',
    'aww',
)
bbc_topics = (
    'news/world',
    'news',
    'sport',
    'business',
    'health',
)
fox_news_topics = (
    'us',
    'politics',
    'world',
    'opinion',
)


class Warmup:
    def __init__(self, *forced_urls: str, min_num: int = 2, max_num: int = 3, excluded_domains: Iterable[str] = ()):
        foxes = tuple(f'https://www.foxnews.com/{x}' for x in random.choices(fox_news_topics, k=2))
        bbc = tuple(f'https://www.bbc.com/{x}' for x in random.choices(bbc_topics, k=2))
        wiki = tuple(f'https://en.wikipedia.org/wiki/{x}' for x in random.choices(wikipedia_topics, k=2))
        reddit = tuple(f'https://www.reddit.com/r/{x}' for x in random.choices(subreddits, k=2))
        urls = (
            (
                'https://www.amazon.com/gp/css/order-history?ref_=nav_orders_first',
                'https://x.com',
                'https://www.youtube.com',
            )
            + foxes
            + bbc
            + wiki
            + reddit
        )
        disallow = frozenset(excluded_domains)
        if disallow:

            def is_not_excluded(x: str) -> bool:
                p = urlparse(x)
                for q in disallow:
                    assert p.hostname is not None
                    if p.hostname.endswith(q):
                        return False
                return True

            urls = tuple(filter(is_not_excluded, urls))
        self.urls = tuple(random.sample(urls, k=random.randint(min_num, max_num))) + forced_urls

    def __call__(self, br: Browser) -> None:
        pass


class Browser(AsyncCamoufox):
    def __init__(self, headless: bool = True, warmup: Warmup | None = None):
        os = 'windows' if iswindows else ('macos' if ismacos else 'linux')
        super().__init__(headless=headless, os=os, humanize=True)
